#include <iostream>
#include <math.h>
#include "Comm.h"
#include "Graph.h"
#include "StackGraph.h"
#include "StackGraphItem.h"
#include "NodeArray.h"
#include "NodeStack.h"

using namespace std;

bool isBipartit(bool *potent);
bool inner_solve();
void solve();
void comm_resolve();
void printBestSolution();

Comm *comm = NULL;

Graph *graph = NULL;
StackGraph *stack = new StackGraph();

NodeArray *n_array = NULL;
NodeStack *n_stack = new NodeStack();
int *neighbours = NULL;

bool *bestSolution = NULL;
int solEdges, solProcess;

bool isBipartit(bool *potent) {
  n_array->reset();
  n_stack->reset();
  
  int fresh = n_array->getFresh();
  int color = 1;
  
  while(fresh != NodeArray::NOT_FOUND) {
    n_array->setColor(fresh, color);
    n_stack->push(fresh);
    
    while(!n_stack->empty()) {
      int aux = n_stack->pop();
      graph->getNeighbours(aux, potent, neighbours);
      color = 3 - n_array->getColor(aux);
      int i = 0;
      
      while(neighbours[i] != Graph::NEIGH_EMPTY) {
        if(n_array->getColor(neighbours[i]) == NodeArray::FRESH) {
          n_array->setColor(neighbours[i], color);
          n_stack->push(neighbours[i]);
        }
        else if(n_array->getColor(neighbours[i]) != color) return false;
        
        i++;
      }
    }

    fresh = n_array->getFresh();
  }

  return true;
}

bool inner_solve() {
  StackGraphItem *item = stack->pop();
  bool *myGraph = item->getGraph();
  int itemEdges = item->getEdges();
  bool res = false;

  if(itemEdges > solEdges) {
    if(isBipartit(myGraph)) {
      delete [] bestSolution;
      bestSolution = myGraph;
      solEdges = itemEdges;
      solProcess = comm->getRank();

      res = true;
    }
    else {
      int first = item->getLastChange() + 1;
      int edges = graph->getEdgesCount();

      for(int i = first; i < edges; i++) {
        bool *potent = item->getGraph();
        potent[i] = false;
        stack->push(potent, itemEdges - 1, edges);
      }

      delete [] myGraph;
    }
  }

  delete item;

  return res;
}

void solve() {
  int i = 1;
  int j = 1;

  comm->start();

  while(comm->working()) {
    while(!stack->empty()) {
      comm->setState(Comm::BUSY);

      if(inner_solve()) comm->sendSolution(solEdges);

      while((i == 0) && comm->ready()) comm_resolve();
      i++;
      i %= 10;
    }

    comm->setState(Comm::WAITING);

    if(!comm->inquiryRunning() && comm->getRank() == 0) {
      comm->sendToken(1, Comm::WAITING);
      if(comm->getProcessesCount() == 1) comm->stop();
    }

    if(j == 0) comm->requestJob();
    j++;
    j %= 80;

    while(comm->ready()) comm_resolve();
  }
}

void comm_resolve() {
  int msgSource = comm->getMessageSource();
  int myRank = comm->getRank();
  int newSol;
  int rcvdToken;

  switch(comm->getMessageType()) {
    case Comm::SOLUTION:
      newSol = comm->receiveSolution();

      if(newSol > solEdges) {
	solEdges = newSol;
	solProcess = msgSource;
      }

      break;

    case Comm::JOB_REQ:
      comm->ackJob();

      StackGraph *send; send = stack->getBottom();
      int length; length = send->getLength();
      StackGraphItem **s; s = new StackGraphItem *[length];
      for(int i = 0; i < length; i++) s[i] = send->pop();

      comm->sendStack(msgSource, s, length, graph->getEdgesCount());

      for(int i = 0; i < length; i++) delete s[i];
      delete [] s;
      delete send;
      break;

    case Comm::STACK:
      if(stack->empty()) comm->receiveJob(stack);
      else cout << "Proces " << myRank << ": Prisla nevyzadana prace :)" << endl;
      break;

    case Comm::TOKEN:
      rcvdToken = comm->receiveToken();

      if(myRank == 0) {
	if(rcvdToken == Comm::WAITING) {
	  comm->sendTerminate(solProcess);
	  comm->stop();
	}
	else comm->closeInquiry(); // muze poslat dalsi WAITING token
      }
      else {
	if(comm->getState() == Comm::BUSY) rcvdToken = Comm::BUSY; // pokud nejsem BUSY, tak token pouze preposlu
	comm->sendToken((myRank + 1) % comm->getProcessesCount(), rcvdToken); // poslu sousedovi napravo
      }

      break;

    case Comm::TERMINATE:
      comm->stop();
      comm->acceptTerminate(bestSolution, graph->getEdgesCount());
      break;

    case Comm::CONFIG:
      comm->receiveConfig(&bestSolution);
      break;

    default:
      cout << "Proces " << myRank << ": Prisel neznamy typ zpravy" << endl;
      break;
  }
}

void printBestSolution() {
  cout << "Proces " << comm->getRank() << ": ";
  cout << "Reseni (od procesu " << solProcess << "):" << endl;
  cout << "  Pocet hran: " << solEdges << endl;
  cout << "  Konfigurace: ";

  int edges = graph->getEdgesCount();
  
  for(int i = 0; i < edges; i++)
    cout << bestSolution[i] << " ";

  cout << endl << "  Vysledny graf: " << endl;
  cout << flush;

  graph->print(bestSolution);
  
  cout << endl << "Celkovy cas reseni: " << comm->getTotalTime() << endl;
  cout << flush;
}

int main(int argc, char *argv[]) {
  comm = Comm::getInstance(argc, argv);

  cout << "Proces " << comm->getRank() << ": Start" << endl;
  cout << flush;

  // P0 nacte graf, provede prvni expanzi a rozdeli zasobnik procesorum
  if(comm->getRank() == 0) {
    // inicializace struktur
    graph = new Graph(argv[1]);
    int edges = graph->getEdgesCount();

    n_array = new NodeArray(graph->getNodesCount());
    neighbours = new int [graph->getNodesCount() + 1];

    bestSolution = new bool [edges];
    for(int i = 0; i < edges; i++) bestSolution[i] = false;

    solEdges = 0;
    solProcess = 0;

    // prvni expanze
    bool *first = new bool [edges];
    for(int i = 0; i < edges; i++) first[i] = true;

    stack->push(first, edges, edges);

    if(comm->getProcessesCount() > 1) {
      inner_solve();
    
      // rozeslani grafu...
      comm->sendGraph(graph);
    
      // ... a zasobniku
      StackGraphItem ***array;
      int width = comm->getProcessesCount();
      int height = (int) ceil((float) edges / width);

      array = new StackGraphItem **[width];

      for(int i = 0; i < width; i++)
        array[i] = new StackGraphItem *[height];

      cout << "Proces 0: Stav zasobniku: " << stack->getLength() << endl;
      cout << flush;

      for(int i = 0; i < height; i++) // naplneni pole
        for(int j = 0; j < width; j++) {
          if(stack->empty()) array[j][i] = NULL;
          else {
	    array[j][i] = stack->pop();
	    array[j][i]->printGraph();
          }
        }

      for(int i = 1; i < width; i++) { // rozesilani zasobniku
        if(array[i][height - 1] == NULL)
          comm->sendStack(i, array[i], height - 1, edges);
        else comm->sendStack(i, array[i], height, edges);

        for(int j = 0; j < height; j++) delete array[i][j];
        delete [] array[i];
      }

      for(int i = 0; i < height; i++) // neco si P0 necha
        stack->push(array[0][i]);

      delete [] array[0];
      delete [] array;
    }

    cout << "Proces 0: Init OK" << endl;
    cout << flush;

    solve();
  }
  else {
    // prijem grafu a zasobniku
    graph = comm->receiveGraph();
    int edges = graph->getEdgesCount();

    n_array = new NodeArray(graph->getNodesCount());
    neighbours = new int [graph->getNodesCount() + 1];

    bestSolution = new bool [edges];
    for(int i = 0; i < edges; i++) bestSolution[i] = false;

    solEdges = 0;
    solProcess = comm->getRank();
  
    comm->receiveStack(stack);

    cout << "Proces " << comm->getRank() << ": Init OK" << endl;
    cout << flush;

    solve();
  }
  
  if(comm->getRank() == 0) printBestSolution();
  
  cout << "Proces " << comm->getRank() << ": Konec" << endl;
  cout << flush;

  return 0;
}

